//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/SampleDesigner/CoreAndShellForm.cpp
//! @brief     Implements class CoreAndShellForm
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/SampleDesigner/CoreAndShellForm.h"
#include "GUI/Model/CatSample/FormFactorItemCatalog.h"
#include "GUI/Model/Sample/CoreAndShellItem.h"
#include "GUI/Model/Sample/FormFactorItems.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Util/ActionFactory.h"
#include "GUI/View/SampleDesigner/FormLayouter.h"
#include "GUI/View/SampleDesigner/SampleEditorController.h"
#include "GUI/View/SampleDesigner/SelectionContainerForm.h"
#include "GUI/View/Tool/GroupBoxCollapser.h"
#include <QAction>
#include <QComboBox>
#include <memory>

namespace {

QComboBox* createFormFactorCombo(QWidget* parent, FormFactorItem* current)
{
    ASSERT(current);
    auto* combo = new QComboBox(parent);
    WheelEventEater::install(combo);
    combo->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);

    for (const auto type : FormFactorItemCatalog::types()) {
        const auto ui = FormFactorItemCatalog::uiInfo(type);
        combo->addItem(QIcon(ui.iconPath), ui.menuEntry, (uint8_t)type);
    }
    combo->setMaxVisibleItems(combo->count());
    combo->setCurrentIndex(combo->findData((uint8_t)FormFactorItemCatalog::type(current)));

    return combo;
}

} // namespace


CoreAndShellForm::CoreAndShellForm(QWidget* parent, CoreAndShellItem* item,
                                   SampleEditorController* ec, bool allowRemove)
    : QGroupBox(parent)
    , m_item(item)
    , m_ec(ec)
{
    setTitle("Sim/shell particle");
    FormLayouter layouter(this, ec);
    layouter.setContentsMargins(30, 6, 0, 0);
    layouter.addVector(item->position(), false);
    layouter.addSelection(item->rotation());
    layouter.addValue(item->abundance());

    // - core
    {
        auto* coreParticleGroup = new QGroupBox(this);
        coreParticleGroup->setObjectName("Particle");

        core.layouter = std::make_unique<FormLayouter>(coreParticleGroup, ec);

        core.formfactorCombo = createFormFactorCombo(
            coreParticleGroup, item->core() != nullptr ? item->core()->formfactorItem() : nullptr);
        connect(core.formfactorCombo, QOverload<int>::of(&QComboBox::currentIndexChanged), this,
                &CoreAndShellForm::onCoreComboChanged);
        core.layouter->addRow("Form factor:", core.formfactorCombo);
        core.collapser = GroupBoxCollapser::installIntoGroupBox(coreParticleGroup);

        auto* showInRealspaceAction =
            ActionFactory::createShowInRealspaceAction(this, "core particle");
        connect(showInRealspaceAction, &QAction::triggered, this,
                &CoreAndShellForm::showCoreInRealspace);
        core.collapser->addAction(showInRealspaceAction);

        createCoreWidgets();

        layouter.addRow(coreParticleGroup);
    }

    // - shell
    {
        auto* shellParticleGroup = new QGroupBox(this);
        shellParticleGroup->setObjectName("Particle");
        shell.layouter = std::make_unique<FormLayouter>(shellParticleGroup, ec);
        shell.formfactorCombo = createFormFactorCombo(
            shellParticleGroup,
            item->shell() != nullptr ? item->shell()->formfactorItem() : nullptr);
        connect(shell.formfactorCombo, QOverload<int>::of(&QComboBox::currentIndexChanged), this,
                &CoreAndShellForm::onShellComboChanged);
        shell.layouter->addRow("Form factor:", shell.formfactorCombo);
        shell.collapser = GroupBoxCollapser::installIntoGroupBox(shellParticleGroup);

        auto* showInRealspaceAction =
            ActionFactory::createShowInRealspaceAction(this, "shell particle");
        connect(showInRealspaceAction, &QAction::triggered, this,
                &CoreAndShellForm::showShellInRealspace);
        shell.collapser->addAction(showInRealspaceAction);

        createShellWidgets();

        layouter.addRow(shellParticleGroup);
    }

    auto* mainCollapser = GroupBoxCollapser::installIntoGroupBox(this);

    auto* showInRealspaceAction = ActionFactory::createShowInRealspaceAction(
        this, "core/shell particle", [=] { ec->requestViewInRealspace(item); });

    mainCollapser->addAction(showInRealspaceAction);

    if (allowRemove) {
        m_removeAction = ActionFactory::createRemoveAction(this, "core/shell particle",
                                                           [=] { ec->removeParticle(item); });
        mainCollapser->addAction(m_removeAction);
    }
}

void CoreAndShellForm::onCoreComboChanged()
{
    while (core.layouter->layout()->rowCount() > 1)
        core.layouter->layout()->removeRow(1);

    const auto type = (FormFactorItemCatalog::Type)core.formfactorCombo->currentData().toUInt();
    m_ec->setCoreFormFactor(this, type);
}

void CoreAndShellForm::onShellComboChanged()
{
    while (shell.layouter->layout()->rowCount() > 1)
        shell.layouter->layout()->removeRow(1);

    const auto type = (FormFactorItemCatalog::Type)shell.formfactorCombo->currentData().toUInt();
    m_ec->setShellFormFactor(this, type);
}

void CoreAndShellForm::showCoreInRealspace()
{
    if (m_item->core())
        m_ec->requestViewInRealspace(m_item->core());
}

void CoreAndShellForm::showShellInRealspace()
{
    if (m_item->shell())
        m_ec->requestViewInRealspace(m_item->shell());
}

void CoreAndShellForm::createCoreWidgets()
{
    QString groupTitle = "Core";

    if (ParticleItem* particle = m_item->core()) {
        const QString formfactor = FormFactorItemCatalog::menuEntry(particle->formfactorItem());
        groupTitle += " (" + formfactor + ")";

        core.layouter->addGroupOfValues("Geometry",
                                        particle->formfactorItem()->geometryProperties());
        core.layouter->addVector(particle->position(), false);
        core.layouter->addSelection(particle->rotation());
        // no abundance since this is handled in CoreShell itself!
    }

    core.collapser->setTitle(groupTitle);
}

void CoreAndShellForm::createShellWidgets()
{
    QString groupTitle = "Shell";

    if (ParticleItem* particle = m_item->shell()) {
        const QString formfactor = FormFactorItemCatalog::menuEntry(particle->formfactorItem());
        groupTitle += " (" + formfactor + ")";

        shell.layouter->addGroupOfValues("Geometry",
                                         particle->formfactorItem()->geometryProperties());
        shell.layouter->addSelection(particle->rotation());
        // no position vector - not allowed in CoreShell
        // no abundance since this is handled in CoreShell itself!
    }

    shell.collapser->setTitle(groupTitle);
}

void CoreAndShellForm::enableStructureEditing(bool b)
{
    if (m_removeAction)
        m_removeAction->setVisible(b);
}

CoreAndShellItem* CoreAndShellForm::coreShellItem() const
{
    return m_item;
}
